/*
**    TP CPE Lyon
**    Copyright (C) 2013 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef MATRIX3_HPP
#define MATRIX3_HPP

#include <iostream>
#include <vector>

#include <exception_cpe.hpp>

namespace cpe
{
class vec3;


/** \brief Matrix 3x3 */
class matrix3
{

public:

    // ********************************************* //
    // ********************************************* //
    //  CONSTRUCTORS
    // ********************************************* //
    // ********************************************* //

    /** \brief empty constructor (identity) */
    matrix3();
    /** \brief direct constructor with every value */
    matrix3(float x00,float x01,float x02,
            float x10,float x11,float x12,
            float x20,float x21,float x22);

    // ********************************************* //
    // ********************************************* //
    //  Special initialization
    // ********************************************* //
    // ********************************************* //

    /** \brief build identity matrix */
    static matrix3 identity();
    /** \brief build zero matrix */
    static matrix3 zeros();
    /** \brief build rotation matrix */
    static matrix3 rotation(const vec3& axis,float angle);
    /** \brief build scaling matrix */
    static matrix3 scale(float s);
    /** \brief build scaling matrix */
    static matrix3 scale(float s_x,float s_y,float s_z);


    // ********************************************* //
    // ********************************************* //
    //  Element access
    // ********************************************* //
    // ********************************************* //

    /** \brief Access to the k_th entry (k in [0,2]) */
    float operator()(const size_t& k1,const size_t& k2) const;
    /** \brief Access to the k_th entry (k in [0,2]) */
    float& operator()(const size_t& k1,const size_t& k2);

    /** \brief fast pointer access
            \warning unprotected access */
    const float* pointer() const;
    /** \brief fast pointer access
            \warning unprotected access */
    float* pointer_unprotected();



    // ********************************************* //
    // ********************************************* //
    //  Math operator
    // ********************************************* //
    // ********************************************* //

    /** \brief + operator */
    matrix3 operator+(const matrix3& m2) const;
    /** \brief + operator */
    matrix3 operator+(float s) const;

    /** \brief - operator */
    matrix3 operator-(const matrix3& m2) const;
    /** \brief - operator */
    matrix3 operator-(float s) const;


    /** \brief multiply by a scalar operator */
    matrix3 operator*(float s) const;
    /** \brief multiply matrix with a v3 */
    vec3 operator*(const vec3& v) const;
    /** \brief matrix mulitplication */
    matrix3 operator*(const matrix3& m2) const;
    /** \brief divide by a scalar operator */
    matrix3 operator/(float s) const;


    /** \brief internal + */
    matrix3& operator+=(const matrix3& m);
    /** \brief internal + */
    matrix3& operator+=(float s);
    /** \brief internal - */
    matrix3& operator-=(const matrix3& m);
    /** \brief internal - */
    matrix3& operator-=(float s);
    /** \brief internal * */
    matrix3& operator*=(float s);
    /** \brief internal / */
    matrix3& operator/=(float s);

    /** \brief unary negation */
    matrix3 operator-() const;

    /** \brief does componentwise multiplication  */
    matrix3 product_compontentwise(const matrix3& m) const;
    /** \brief does componentwise multiplication  */
    matrix3& product_compontentwise_internal(const matrix3& m);

    /** \brief transpose matrix */
    matrix3 transposed() const;



    // ********************************************* //
    // ********************************************* //
    //  Projection matrix
    // ********************************************* //
    // ********************************************* //

    /** \brief convert 3x3 matrix into a vector of float of size 9 */
    std::vector<float> to_vector() const;

private:

    // ********************************************* //
    // ********************************************* //
    //  Internal parameters
    // ********************************************* //
    // ********************************************* //

    /** \brief internal storage of the matrix */
    float m[9];


    // ********************************************* //
    // ********************************************* //
    //  Helper
    // ********************************************* //
    // ********************************************* //

    /** \brief assert that a size_t belongs to [[0,2]] */
    void assert_size(const size_t& k1,const size_t& k2) const;

};

/** \brief + operator */
matrix3 operator+(float s,const matrix3& m);
/** \brief - operator */
matrix3 operator-(float s,const matrix3& m);
/** \brief multiply by a scalar operator */
matrix3 operator*(float s,const matrix3& m);

// ********************************************* //
// ********************************************* //
//  Output
// ********************************************* //
// ********************************************* //

/** \brief output the vector in ostream */
std::ostream& operator<<(std::ostream& stream,const matrix3& m);

/** \brief Special exception class for v3 */
class exception_matrix3 : public exception_cpe
{
public:

    /** \brief empty constructor */
    exception_matrix3():exception_cpe(){}
    /** \brief constructor with error information */
    exception_matrix3(const std::string& msg_param,const std::string& file_param,const std::string& caller_param,const int& line_param):exception_cpe(msg_param,file_param,caller_param,line_param){}
};



}

#endif
