/*
**    TP CPE Lyon
**    Copyright (C) 2013 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#include <vec3.hpp>

#include <vec2.hpp>
#include <cmath>
#include <sstream>

#include <assert.h>

namespace cpe
{

    vec3::vec3():internal_x(0),internal_y(0),internal_z(0){}
    vec3::vec3(float x_param,float y_param,float z_param):internal_x(x_param),internal_y(y_param),internal_z(z_param){}
    float vec3::x() const
    {
        return internal_x;
    }

    float& vec3::x()
    {
        return internal_x;
    }

    float vec3::y() const
    {
        return internal_y;
    }

    float& vec3::y()
    {
        return internal_y;
    }

    float vec3::z() const
    {
        return internal_z;
    }

    float& vec3::z()
    {
        return internal_z;
    }

    float vec3::operator[](const size_t& k) const
    {
        assert_size(k);
	switch (k)
	  {
	  case 0:
	    return internal_x;
	  case 1:
	    return internal_y;
	  case 2:
	    return internal_z;
	  default:
	    assert(false);
        throw cpe::exception_vec3("Probleme",__FILE__,__FUNCTION__,__LINE__);
	  };
    }

    float& vec3::operator[](const size_t& k)
    {
        assert_size(k);
	switch (k)
	  {
	  case 0:
	    return internal_x;
	  case 1:
	    return internal_y;
	  case 2:
	    return internal_z;
	  default:
	    assert(false);
        throw cpe::exception_vec3("Probleme",__FILE__,__FUNCTION__,__LINE__);
	  };
    }

    float vec3::operator()(const size_t& k) const
    {
      return (*this)[k];
    }
    float& vec3::operator()(const size_t& k)
    {
      return (*this)[k];
    }

    void vec3::set_zero()
    {
        internal_x=0;
        internal_y=0;
        internal_z=0;
    }

    float vec3::dot(const vec3& p) const
    {
        return internal_x*p.x()+internal_y*p.y()+internal_z*p.z();
    }

    float vec3::norm() const
    {
        return std::sqrt(norm2());
    }

    float vec3::norm2() const
    {
        return internal_x*internal_x+internal_y*internal_y+internal_z*internal_z;
    }

    vec3 vec3::cross(const vec3& p) const
    {
        return vec3(internal_y*p.z()-internal_z*p.y(),
                  internal_z*p.x()-internal_x*p.z(),
                  internal_x*p.y()-internal_y*p.x()
                  );
    }
    vec3 vec3::operator+(const vec3& p2) const
    {
        return vec3(x()+p2.x(),y()+p2.y(),z()+p2.z());
    }

    vec3 operator+(float s,const vec3& p)
    {
        return vec3(s+p.x(),s+p.y(),s+p.z());
    }

    vec3 vec3::operator+(float s) const
    {
        vec3 temp=*this;
        temp+=s;
        return temp;
    }

    vec3 vec3::operator-(const vec3& p2) const
    {
        vec3 temp=*this;
        temp-=p2;
        return temp;
    }

    vec3 operator-(float s,const vec3& p)
    {
        return vec3(s-p.x(),s-p.y(),s-p.z());
    }

    vec3 vec3::operator-(float s) const
    {
        return vec3(x()-s,y()-s,z()-s);
    }

    vec3 operator*(float s,const vec3& p)
    {
        return vec3(p.x()*s,p.y()*s,p.z()*s);
    }

    vec3 vec3::operator*(float s) const
    {
        vec3 temp=*this;
        temp*=s;
        return temp;
    }

    vec3 vec3::operator/(float s) const
    {
      float epsilon=1e-8;
      if(std::abs(s)<epsilon)
          throw cpe::exception_vec3("Division par 0",__FILE__,__FUNCTION__,__LINE__);

      return vec3(x()/s,y()/s,z()/s);
    }

    vec3& vec3::operator+=(const vec3& p)
    {
        internal_x+=p.x();
        internal_y+=p.y();
        internal_z+=p.z();
        return *this;
    }

    vec3& vec3::operator+=(float s)
    {
        internal_x+=s;
        internal_y+=s;
        internal_z+=s;
        return *this;
    }

    vec3& vec3::operator-=(const vec3& p)
    {
        internal_x-=p.x();
        internal_y-=p.y();
        internal_z-=p.z();
        return *this;
    }

    vec3& vec3::operator-=(float s)
    {
        internal_x-=s;
        internal_y-=s;
        internal_z-=s;
        return *this;
    }

    vec3& vec3::operator*=(float s)
    {
        internal_x*=s;
        internal_y*=s;
        internal_z*=s;
        return *this;
    }

    vec3& vec3::operator/=(float s)
    {
        internal_x/=s;
        internal_y/=s;
        internal_z/=s;
        return *this;
    }

    vec3 vec3::operator-() const
    {
        return vec3(-internal_x,-internal_y,-internal_z);
    }

    vec3 vec3::product_compontentwise(const vec3& p) const
    {
        return vec3(internal_x*p.x(),internal_y*p.y(),internal_z*p.z());
    }

    vec3& vec3::product_compontentwise_internal(const vec3& p)
    {
        internal_x*=p.x();
        internal_y*=p.y();
        internal_z*=p.z();
        return *this;
    }
    std::string vec3::to_string() const
    {
        std::stringstream stream;
        stream<<internal_x<<" "<<internal_y<<" "<<internal_z;
        return stream.str();
    }
    std::ostream& operator<<(std::ostream& stream,const vec3& p)
    {
        stream<<"("<<p.to_string()<<")";
        return stream;
    }

    void vec3::assert_size(const size_t& k) const
    {
        if(k>2)
        {
            std::stringstream stream;
            stream<<"vec3::assert_size("<<k<<") k>2";
            throw exception_vec3(stream.str(),__FILE__,__FUNCTION__,__LINE__);
        }
    }
    vec3 vec3::normalized() const
    {
        float epsilon=1e-6;

        float n=norm();
        if(n<epsilon)
            throw exception_vec3("norm=0",__FILE__,__FUNCTION__,__LINE__);
        return *this/n;
    }
    void vec3::scale(float sx,float sy,float sz)
    {
        internal_x*=sx;
        internal_y*=sy;
        internal_z*=sz;
    }

    vec2 vec3::to_vec2() const{return vec2(internal_x,internal_y);}

    const float *vec3::pointer() const
    {return &internal_x;}
}
