/*
**    TP CPE Lyon
**    Copyright (C) 2013 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef Z_BUFFER_HPP
#define Z_BUFFER_HPP

#include <vector>
#include <string>
#include <pos2.hpp>

namespace cpe
{

/** \brief Class to manage a depth-buffer */
class z_buffer
{
    public:

        // ********************************************* //
        // ********************************************* //
        //  CONSTRUCTORS
        // ********************************************* //
        // ********************************************* //

        /** \brief Constructor with size=0 */
        z_buffer();
        /** \brief Constructor with square picture of size: size*size
            \note Init the z-buffer to infinity
         */
        z_buffer(int size);
        /** \brief Constructor with square picture of given size
            \note Init the z-buffer to infinity
         */
        z_buffer(int size_x,int size_y);


        // ********************************************* //
        // ********************************************* //
        //  Manipulation
        // ********************************************* //
        // ********************************************* //

        /** \brief fill the z-buffe with special value */
        void fill(double value);

        /** \brief Accessor of value of the z-buffer */
        double operator()(const pos2& u) const;
        /** \brief Accessor of value of the z-buffer */
        double& operator()(const pos2& u);

        // ********************************************* //
        // ********************************************* //
        //  Info
        // ********************************************* //
        // ********************************************* //

        /** \brief return the max value of the z-buffer (eliminating the infinity value) */
        double get_max() const;
        /** \brief return the min value of the z-buffer */
        double get_min() const;

        // ********************************************* //
        // ********************************************* //
        //  Size
        // ********************************************* //
        // ********************************************* //

        /** \brief size of the z-buffer in x */
        int size_x() const;
        /** \brief size of the z-buffer in y */
        int size_y() const;

        // ********************************************* //
        // ********************************************* //
        //  I/O
        // ********************************************* //
        // ********************************************* //

        /** \brief write the z-buffer as an image into a file */
        void export_file(const std::string& filename) const;


    protected:

        /** \brief convert (x,y) coordinate to index */
        int coordinate_to_index(const pos2& u) const;

        /** \brief check if u is in bounds of the picture */
        void assert_position(const pos2& u) const;

    private:



        /** \brief check internal data  */
        std::vector<double> internal_data;

        /** \brief size x of the z-buffer */
        int internal_size_x;
        /** \brief size y of the z-buffer */
        int internal_size_y;

        /** \brief infinity value */
        static const double depth_infinity=9e5;
};
}

#endif
